package ds;

import java.math.BigInteger;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Random;

import remoteInterfaces.BulletinBoard;
import votingSystem.BinaryVote;
import votingSystem.DecryptedShare;
import votingSystem.ElectionProperties;
import votingSystem.HashFunction;
import votingSystem.KeyShare;
import votingSystem.MultipleBinaryVotes;
import votingSystem.MultipleVotes;
import votingSystem.PublicKey;
import votingSystem.Utility;
import votingSystem.Validator;
import votingSystem.Vote;
import exceptions.NoValidVotesException;

/**
 * This is the class for decryption servers with the main functions for
 * checking proofs on votes as well as share-decrypting votes.
 * 
 * 
 * @author bent
 * 
 */
public class Server {

	private BigInteger vi, si, delta, v;

	private BigInteger n;

	private BigInteger[] voteValues;

	private int k, id, electionType;

	private int s;

	private int M;

	private int t;

	private int L;

	private Random r;

	private HashFunction h;

	private KeyShare ks;

	private PublicKey pk;

	private BigInteger nArray[];

	private int noOfVotes;

	private BigInteger secretExponent;
	
	private BulletinBoard bb;

	private Validator val;

	/**
	 * Constructs a DecryptionServer to decrypt the results using the servers
	 * secret KeyShare
	 * 
	 * @param id
	 *            The DecryptionServer id
	 * @param ks
	 *            The secret KeyShare
	 * @param ep
	 *            The ElectionProperties for this election
	 * @param r
	 *            The random-number generator
	 * @param pk
	 *            The PublicKey for the election
	 */
	public Server(int id, KeyShare ks, ElectionProperties ep, Random r,
			PublicKey pk, BulletinBoard bb) {
		this.ks = ks;
		this.id = id;
		v = ks.getV();
		vi = ks.getVi();
		si = ks.getSi();
		val = new Validator(ep, pk);
		this.t = ep.getT();
		this.delta = BigInteger.valueOf(Utility.faculty(ep.getl()));
		this.n = ks.getN();
		this.M = ep.getM();
		this.s = ep.getS();
		this.k = ep.getK();
		this.r = r;
		this.L = ep.getL();
		this.secretExponent = (delta.multiply(si)).multiply(BigInteger.valueOf(2));
		this.voteValues = ep.getVoteValues();
		this.noOfVotes = ep.getNoOfVotes();
		this.pk = pk;
		this.electionType = ep.getElectionType();
		this.bb = bb;
		h = new HashFunction("SHA-256");
		nArray = pk.getNArray();

	}

	/**
	 * Constructs a DecryptionServer to decrypt the results using the servers
	 * secret KeyShare
	 * 
	 * @param id
	 *            The DecryptionServer id
	 * @param ks
	 *            The secret KeyShare
	 * @param ep
	 *            The ElectionProperties for this election
	 * @param r
	 *            The random-number generator
	 * @param pk
	 *            The PublicKey for the election
	 */
	public Server(int id, KeyShare ks, ElectionProperties ep, Random r,
			PublicKey pk) {
		this.ks = ks;
		this.id = id;
		v = ks.getV();
		vi = ks.getVi();
		si = ks.getSi();
		this.t = ep.getT();
		this.delta = BigInteger.valueOf(Utility.faculty(ep.getl()));
		this.n = ks.getN();
		this.M = ep.getM();
		this.s = ep.getS();
		this.k = ep.getK();
		val = new Validator(ep, pk);
		this.r = r;
		this.L = ep.getL();
		this.voteValues = ep.getVoteValues();
		this.noOfVotes = ep.getNoOfVotes();
		this.electionType = ep.getElectionType();
		this.pk = pk;
		h = new HashFunction("SHA-256");
		nArray = pk.getNArray();

	}
/** Initiates the process of share decrypting the results of the election.
 * 
 * @throws RemoteException
 * @throws NoValidVotesException
 */
	public void postShare() throws RemoteException, NoValidVotesException {
		System.out.println("Posting Share");
		switch (electionType) {
		case 0:
			DecryptedShare dshare = shareDecrypt(bb.getVotes());

			Vote[] v = bb.getVotes();
			bb.postShare(dshare);
			break;

		case 1:
			DecryptedShare[] dshareList = shareDecrypt(bb.getMultipleVotes());
			bb.postShare(dshareList);
			break;
		case 2:
			DecryptedShare dshareBinary = shareDecrypt(bb.getBinaryVotes());
			bb.postShare(dshareBinary);
			break;
		case 3:
			DecryptedShare[] dShareBinaryList = shareDecrypt(bb
					.getMultipleBinaryVotes());
			bb.postShare(dShareBinaryList);
			break;
		}

	}

	/**
	 * The method for decrypting the results of the election with the secret
	 * KeyShare
	 * 
	 * @param votes
	 *            The votes from the election
	 */
	public DecryptedShare[] shareDecrypt(MultipleVotes[] votes)
			throws NoValidVotesException {
		int bitlength = (s + 2) * k + t;
		int validVotesNo = 0;
		ArrayList validVotes = new ArrayList();
		for (int i = 0; i < votes.length; i++) {
			if (this.val.checkMultipleVote(votes[i]))
				validVotes.add(votes[i]);
		}

		
		if (validVotes.size() == 0)
			throw new NoValidVotesException("No valid votes given");

		BigInteger c,ci, temp, random, a, b,e,z,result;
		BigInteger[] values = new BigInteger[5];
		values[0] = n;
		DecryptedShare[] shares = new DecryptedShare[noOfVotes];
		for (int k = 0; k < noOfVotes; k++) {
			result = ((MultipleVotes)validVotes.get(0)).getVotes()[k].getVote();
			for (int i = 1; i < validVotes.size(); i++) {
				result = result.multiply(((MultipleVotes)validVotes.get(i)).getVotes()[k].getVote())
						.mod(nArray[s + 1]);
			}
			 c = result;
			 //temp = c.modPow(delta.multiply(BigInteger.valueOf(2)),
			//		nArray[s + 1]);
			 ci = c.modPow(secretExponent, nArray[s + 1]);

			 random = new BigInteger(bitlength, r);
			// Generate proof!!
			 a = (c.pow(4)).mod(nArray[s + 1]);
			a = a.modPow(random, nArray[s + 1]);

			 b = v.modPow(random, nArray[s + 1]);

			// Random oracle!
			values[1] = a;
			values[2] = b;
			values[3] = c.pow(4).mod(nArray[s + 1]);
			values[4] = ci.pow(2).mod(nArray[s + 1]);

			e = h.generateHash(values, t);
			temp = (e.multiply(si).multiply(delta));
			 z = random.add(temp);


			shares[k] = new DecryptedShare(e, z, c, ci, id);
		}
		return shares;
	}

	/**
	 * The method for decrypting the results of the election with the secret
	 * KeyShare
	 * 
	 * @param votes
	 *            The votes from the election
	 */
	public DecryptedShare shareDecrypt(Vote[] votes)
			throws NoValidVotesException {
		int bitlength = (s + 2) * k + t;
		ArrayList validVotes = new ArrayList();

		for (int i = 0; i < votes.length; i++) {
			if (val.checkProof(votes[i]))
				validVotes.add(votes[i]);
		}

		if (validVotes.size() == 0)
			throw new NoValidVotesException("No valid votes given");
		//Calculate encrypted result of the election
		BigInteger result = ((Vote) validVotes.get(0)).getVote();
		for (int i = 1; i < validVotes.size(); i++) {
			result = result.multiply(((Vote) validVotes.get(i)).getVote()).mod(
					nArray[s + 1]);
		}
		BigInteger c = result;
		BigInteger ci = c.modPow(secretExponent, nArray[s + 1]);

		BigInteger random = new BigInteger(bitlength, r);
		// Generate zero-knowledge proof
		BigInteger a = (c.pow(4)).mod(nArray[s + 1]);
		a = a.modPow(random, nArray[s + 1]);

		BigInteger b = v.modPow(random, nArray[s + 1]);

		BigInteger e;
		BigInteger[] values = new BigInteger[5];
		values[0] = n;
		values[1] = a;
		values[2] = b;
		values[3] = c.pow(4).mod(nArray[s + 1]);
		values[4] = ci.pow(2).mod(nArray[s + 1]);

		e = h.generateHash(values, t);
		BigInteger temp2 = (e.multiply(si).multiply(delta));
		BigInteger z = random.add(temp2);

		DecryptedShare ds = new DecryptedShare(e, z, c, ci, id);
		return ds;
	}
	/**
	 * The method for decrypting the results of the election with the secret
	 * KeyShare
	 * 
	 * @param votes
	 *            The votes from the election
	 */
	public DecryptedShare[] shareDecrypt(MultipleBinaryVotes[] votes)
			throws NoValidVotesException {
	
		int bitlength = (s + 2) * k + t;
		ArrayList validVotes = new ArrayList();
		
		for (int i = 0; i < votes.length; i++) {
			if (val.checkMultipleBinaryVote(votes[i]))
				validVotes.add(votes[i]);
		}
		if (validVotes.size() == 0)
			throw new NoValidVotesException("No valid votes given");
		
		DecryptedShare[] dsList = new DecryptedShare[noOfVotes];
		
		BigInteger[] values = new BigInteger[5];
		values[0] = n;
		BigInteger c, tempval, ci, random, a, b,e;

		int l = ((MultipleBinaryVotes)validVotes.get(0)).getVotes()[0].getFi().length-1;
		for (int i = 0; i < noOfVotes; i++) {
			 c = ((MultipleBinaryVotes) validVotes.get(0)).getVotes()[i].getFi()[l];
			for (int j = 1; j < validVotes.size(); j++) {
				c = c.multiply(((MultipleBinaryVotes) validVotes.get(j)).getVotes()[i].getFi()[l]).mod(nArray[s+1]);
			}
			//tempval = c.modPow(delta.multiply(BigInteger.valueOf(2)),nArray[s + 1]);
			ci = c.modPow(secretExponent, nArray[s + 1]);

			random = new BigInteger(bitlength, r);
			// Generate proof!!
			a = (c.pow(4)).mod(nArray[s + 1]);
			a = a.modPow(random, nArray[s + 1]);
			b = v.modPow(random, nArray[s + 1]);

			values[1] = a;
			values[2] = b;
			values[3] = c.pow(4).mod(nArray[s + 1]);
			values[4] = ci.pow(2).mod(nArray[s + 1]);

			e = h.generateHash(values, t);
			tempval = (e.multiply(si).multiply(delta));
			BigInteger z = random.add(tempval);
			dsList[i] = new DecryptedShare(e, z, c, ci, id);
		}
		return dsList;

	}
	/**
	 * The method for decrypting the results of the election with the secret
	 * KeyShare
	 * 
	 * @param votes
	 *            The votes from the election
	 */
	public DecryptedShare shareDecrypt(BinaryVote[] votes)
			throws NoValidVotesException {
		int bitlength = (s + 2) * k + t;
		ArrayList validVotes = new ArrayList();

		for (int i = 0; i < votes.length; i++) {
			if (val.checkBinaryVote(votes[i]))
				validVotes.add(votes[i]);
		}

		if (validVotes.size() == 0)
			throw new NoValidVotesException("No valid votes given");
		BigInteger[] fi = ((BinaryVote) validVotes.get(0)).getFi();
		int l = fi.length-1;
		
		BigInteger result = fi[l];
		for (int i = 1; i < validVotes.size(); i++) {
			fi = ((BinaryVote) validVotes.get(i)).getFi();
			result = result.multiply(fi[l]).mod(nArray[s + 1]);
		}
		BigInteger c = result;

	//	BigInteger temp = c.modPow(delta.multiply(BigInteger.valueOf(2)),
		//		nArray[s + 1]);
		BigInteger ci = c.modPow(secretExponent, nArray[s + 1]);

		BigInteger random = new BigInteger(bitlength, r);
		// Generate proof!!
		BigInteger a = (c.pow(4)).mod(nArray[s + 1]);
		a = a.modPow(random, nArray[s + 1]);

		BigInteger b = v.modPow(random, nArray[s + 1]);

		// Random oracle!
		BigInteger e;
		BigInteger[] values = new BigInteger[5];
		values[0] = n;
		values[1] = a;
		values[2] = b;
		values[3] = c.pow(4).mod(nArray[s + 1]);
		values[4] = ci.pow(2).mod(nArray[s + 1]);

		e = h.generateHash(values, t);
		BigInteger temp2 = (e.multiply(si).multiply(delta));
		BigInteger z = random.add(temp2);

		DecryptedShare ds = new DecryptedShare(e, z, c, ci, id);
		return ds;

	}

}
