package ea;

import java.math.BigInteger;
import java.rmi.RemoteException;
import java.util.ArrayList;

import remoteInterfaces.BulletinBoard;
import votingSystem.BinaryVote;
import votingSystem.DecryptedShare;
import votingSystem.ElectionProperties;
import votingSystem.HashFunction;
import votingSystem.MultipleBinaryVotes;
import votingSystem.MultipleVotes;
import votingSystem.PublicKey;
import votingSystem.Result;
import votingSystem.Utility;
import votingSystem.Validator;
import votingSystem.Vote;
import exceptions.InsufficientSharesException;

/**
 * This class resembles the ElectionAuthority of the Election. This class will
 * handle the decryption of the results when enough decrypted shares are
 * acquired from the Servers
 * 
 * @author bent
 * 
 */

public class ElectionAuthority {

	private PublicKey pk;

	private ElectionProperties ep;

	private HashFunction h;

	private BigInteger v;

	private BigInteger[] nArray;

	private int s;

	private BulletinBoard bb;

	private Result r;

	private Validator val;

	private boolean trustDecryptionServers;

	/**
	 * Creates a new ElectionAuthority
	 * 
	 * @param pk
	 *            The PublicKey for this election
	 * @param ep
	 *            The ElectionProperties for this election
	 * @param v
	 *            The fixed public value v, that is the generator for the
	 *            verification keys.
	 */
	public ElectionAuthority(PublicKey pk, ElectionProperties ep,
			BulletinBoard bb, boolean trustDecryptionServers) {
		this.bb = bb;
		this.pk = pk;
		this.ep = ep;
		this.trustDecryptionServers = trustDecryptionServers;
		this.v = pk.getV();
		h = new HashFunction("SHA-256");
		nArray = pk.getNArray();
		s = pk.getS();
		r = new Result();
		val = new Validator(ep, pk);
	}

	/**
	 * Creates a new ElectionAuthority
	 * 
	 * @param pk
	 *            The PublicKey for this election
	 * @param ep
	 *            The ElectionProperties for this election
	 * 
	 */
	public ElectionAuthority(PublicKey pk, ElectionProperties ep, boolean trustDecryptionServers) {
		this.pk = pk;
		this.ep = ep;
		this.v = pk.getV();;
		h = new HashFunction("SHA-256");
		nArray = pk.getNArray();
		s = pk.getS();
		val = new Validator(ep, pk);
	}
/** Combines the shares into a result.
 * 
 * @param type The election type. 0 if 1-out of L, 1 if k-out-of-L, 2 if 1-out-of-L Binary, 3 if k-out-of-L Binary.
 * @return The final results of the election
 * @throws RemoteException
 * @throws InsufficientSharesException
 */
	public BigInteger[] combineShares(int type) throws RemoteException,
			InsufficientSharesException {
		System.out.println("ElectionType: " + type);
		switch (type) {
		case 0:
			DecryptedShare[] ds = bb.getOneOfLShares();

			if (!trustDecryptionServers) {
				Vote[] votes = bb.getVotes();

				ArrayList validVotes = new ArrayList();

				for (int i = 0; i < votes.length; i++) {
					if (val.checkProof(votes[i]))
						validVotes.add(votes[i]);
				}

				BigInteger totalVotes = ((Vote) validVotes.get(0)).getVote();
				for (int i = 1; i < validVotes.size(); i++) {
					totalVotes = totalVotes.multiply(
							((Vote) validVotes.get(i)).getVote()).mod(
							nArray[s + 1]);
				}
				ds = checkShareVoteValues(ds, totalVotes);
			}

			BigInteger result = shareCombine(ds);
			return r.getResult(result, ep.getVoteValues());

		case 1:
			ArrayList ds2 = bb.getKOfLShares();

			if (!trustDecryptionServers) {
				MultipleVotes[] votes = bb.getMultipleVotes();

				ArrayList validVotes = new ArrayList();
				for (int i = 0; i < votes.length; i++) {
					if (val.checkMultipleVote(votes[i]))
						validVotes.add(votes[i]);
				}

				for (int j = 0; j < ep.getNoOfVotes(); j++) {
					BigInteger totalVotes = ((MultipleVotes) validVotes.get(0))
							.getVotes()[j].getVote();
					for (int i = 1; i < validVotes.size(); i++) {
						totalVotes = totalVotes
								.multiply(
										((MultipleVotes) validVotes.get(i))
												.getVotes()[j].getVote()).mod(
										nArray[s + 1]);
					}
					for (int i = 0; i < ds2.size(); i++) {
						if ((((DecryptedShare[]) ds2.get(i))[j].getC())
								.compareTo(totalVotes) != 0)
							ds2.remove(i);
					}
				}
			}
			BigInteger[] combinedShare = combineKOfLShares(ds2);

			return r.getResult(combinedShare, ep.getVoteValues(), ep
					.getVoteWeights());

		case 2:
			DecryptedShare[] dsBin = bb.getOneOfLShares();

			if (!trustDecryptionServers) {
				BinaryVote[] binvotes = bb.getBinaryVotes();

				ArrayList validVotes = new ArrayList();

				for (int i = 0; i < binvotes.length; i++) {
					if (val.checkBinaryVote(binvotes[i]))
						validVotes.add(binvotes[i]);
				}
				int l = ((BinaryVote) validVotes.get(0)).getFi().length - 1;
				BigInteger totalVotes_ = ((BinaryVote) validVotes.get(0))
						.getFi()[l];
				for (int i = 1; i < validVotes.size(); i++) {
					totalVotes_ = totalVotes_.multiply(
							((BinaryVote) validVotes.get(i)).getFi()[l]).mod(
							nArray[s + 1]);
				}
				dsBin = checkShareVoteValues(dsBin, totalVotes_);
			}
			BigInteger resultBin = shareCombine(dsBin);
			return r.getResultFromBinary(resultBin, ep.getM(), ep.getL());

		case 3:
			ArrayList dsBinMult = bb.getKOfLShares();
			if (!trustDecryptionServers) {
				MultipleBinaryVotes[] votes = bb.getMultipleBinaryVotes();

				ArrayList validVotes = new ArrayList();
				for (int i = 0; i < votes.length; i++) {
					if (val.checkMultipleBinaryVote(votes[i]))
						validVotes.add(votes[i]);
				}

				for (int j = 0; j < ep.getNoOfVotes(); j++) {
					int l = ((MultipleBinaryVotes) validVotes.get(0))
							.getVotes()[j].getFi().length - 1;
					BigInteger totalVotes = ((MultipleBinaryVotes) validVotes
							.get(0)).getVotes()[j].getFi()[l];
					
					for (int i = 1; i < validVotes.size(); i++) {
						totalVotes = totalVotes.multiply(
								((MultipleBinaryVotes) validVotes.get(i))
										.getVotes()[j].getFi()[l]).mod(
								nArray[s + 1]);
					}
					
					for (int i = 0; i < dsBinMult.size(); i++) {
						if ((((DecryptedShare[]) dsBinMult.get(i))[j].getC())
								.compareTo(totalVotes) != 0){
							dsBinMult.remove(i);
							System.out.println("removed nr: "+i);
						}
					}
				}
			}
			BigInteger[] combinedShare2 = combineKOfLShares(dsBinMult);
			return r.getResultFromBinary(combinedShare2, ep.getM(), ep.getL(),
					ep.getVoteWeights());

		default:
			return null;

		}
	}
/** Checks if the DecryptedShares are calculated from the correct set of valid votes, and removes any DecryptedShares
 * that has calculated the share from a wrong set of votes
 * 
 * @param ds The decryption servers
 * @param voteValues The ciphertext of the result 
 * @return The decryption servers that has used the correct set of votes.
 */
	private DecryptedShare[] checkShareVoteValues(DecryptedShare[] ds,
			BigInteger voteValues) {
		ArrayList correctVoteValueShares = new ArrayList();
		for (int i = 0; i < ds.length; i++) {
			if (ds[i].getC().compareTo(voteValues) == 0) {
				correctVoteValueShares.add(ds[i]);
			}
		}
		return (DecryptedShare[]) correctVoteValueShares
				.toArray(new DecryptedShare[correctVoteValueShares.size()]);
	}
/** Combines the shares in the case of a multi candidate election
 * 
 * @param dsList
 * @return The undecoded results of the election
 * @throws InsufficientSharesException
 */
	private BigInteger[] combineKOfLShares(ArrayList dsList)
			throws InsufficientSharesException {
		System.out.println("In CombineKOfLShares!");
		BigInteger results[] = new BigInteger[ep.getNoOfVotes()];
		for (int i = 0; i < ep.getNoOfVotes(); i++) {
			DecryptedShare[] dsL = new DecryptedShare[dsList.size()];
			for (int j = 0; j < dsL.length; j++) {
				dsL[j] = ((DecryptedShare[]) dsList.get(j))[i];
			}

				results[i] = shareCombine(dsL);
		
		}
		return results;
	}

	/**
	 * The method for combining the DecryptedShare's acquired from the
	 * decryption Servers
	 * 
	 * @param dsList
	 *            Array of DecryptedShare's
	 * @return The result of the election
	 * @throws InsufficientSharesException
	 */
	public BigInteger shareCombine(DecryptedShare[] dsList)
			throws InsufficientSharesException {
		if (dsList.length < ep.getW()) {
			throw new InsufficientSharesException("Not enough shares acquired");
		}
		// Check ShareS!
		BigInteger n = pk.getN();
		int needed = ep.getW();
		int[] goodShares = new int[needed];
		int numberOfGoodShares = 0;
		for (int i = 0; i < dsList.length; i++) {
			if (val.checkShareDecryption(dsList[i])) {
				goodShares[numberOfGoodShares] = i;
				numberOfGoodShares++;
				if (numberOfGoodShares >= needed)
					break;
			}
		}
		if (numberOfGoodShares + 1 < needed) {
			throw new InsufficientSharesException(
					"Not enough good shares acquired");
		}
		BigInteger lambda = BigInteger.ZERO;
		BigInteger c_ = BigInteger.ONE;
		BigInteger delta = BigInteger.valueOf(Utility.faculty(ep.getl()));

		for (int i = 0; i < needed; i++) {
			lambda = BigInteger.ONE;
			for (int j = 0; j < needed; j++) {
				if (j != i) {
					// lambda = lambda.multiply (new BigInteger ("-" +
					// (goodShares[j]+1)));
					lambda = lambda.multiply(new BigInteger("-"
							+ (dsList[goodShares[j]].getTallyId() + 1)));
				}
			}
			lambda = lambda.multiply(delta);
			for (int j = 0; j < needed; j++) {
				if (j != i) {
					lambda = lambda
							.divide((new BigInteger(""
									+ (dsList[goodShares[i]].getTallyId() + 1)))
									.subtract(new BigInteger(""
											+ (dsList[goodShares[j]]
													.getTallyId() + 1))));
				}
			}
			c_ = (c_.multiply(dsList[goodShares[i]].getCi().modPow(
					lambda.multiply(new BigInteger("2")), nArray[s + 1])))
					.mod(nArray[s + 1]);

		}
		// Applying algorithm from Theorem 1!
		int s = ep.getS();

		BigInteger c_temp = c_;
		BigInteger temp;
		BigInteger[] lArray = new BigInteger[s + 1];
		for (int i = 2; i <= s; i++) {
			temp = c_temp.mod(n.pow(i));
			temp = temp.subtract(BigInteger.ONE);
			lArray[i - 1] = temp.divide(n);
		}
		temp = c_temp.subtract(BigInteger.ONE);
		lArray[s] = temp.divide(n);

		BigInteger[] iArray = new BigInteger[s + 1];

		iArray[0] = BigInteger.ZERO;

		BigInteger counter = BigInteger.ONE;
		BigInteger faculty[] = new BigInteger[s];
		temp = BigInteger.ONE;

		for (int i = 1; i < s; i++) {
			counter = counter.add(BigInteger.ONE);
			temp = temp.multiply(counter);
			faculty[i - 1] = (nArray[i].multiply(temp.modInverse(nArray[s])))
					.mod(nArray[s]);
		}

		BigInteger t1, t2;
		BigInteger i = BigInteger.ZERO;
		for (int j = 1; j < s + 1; j++) {
			t1 = lArray[j];
			t2 = i;
			for (int k = 2; k <= j; k++) {
				i = i.subtract(BigInteger.ONE);
				t2 = (t2.multiply(i)).mod(nArray[j]);
				t1 = (t1.subtract(n.pow(k - 1).multiply(
						t2.multiply(BigInteger.valueOf(Utility.faculty(j))
								.modInverse(nArray[j]))))).mod(nArray[j]);
			}
			i = t1;
		}
		temp = delta.multiply(delta).multiply(new BigInteger("4"));
		temp = temp.modInverse(nArray[s]);
		i = (i.multiply(temp)).mod(nArray[s]);

		return i;
	}

}
