package voter;

import java.math.BigInteger;
import java.rmi.RemoteException;
import java.util.Random;

import remoteInterfaces.BulletinBoard;
import votingSystem.BinaryVote;
import votingSystem.ElectionProperties;
import votingSystem.HashFunction;
import votingSystem.MultipleBinaryVotes;
import votingSystem.MultipleVotes;
import votingSystem.PublicKey;
import votingSystem.Vote;
import exceptions.IncorrectParameterException;
import exceptions.IncorrectVoteException;

/**
 * The class for the Client side of the election.
 * 
 * @author bent
 * 
 */
public class Voter {

	private int k, L, M, t, s, id, l;

	private BigInteger n;

	private BigInteger[] voteValues;

	private Random rand;

	private HashFunction h;

	private PublicKey pk;

	private BigInteger[] nArray;

	private BigInteger[] voteRandom;

	private BigInteger[] precomp;

	private int votePointer;

	private int noOfVotes;

	private BulletinBoard bb;

	/**
	 * Creates a new Voter for the election
	 * 
	 * @param id
	 *            The Voter id
	 * @param ep
	 *            The ElectionProperties for the election
	 * @param pk
	 *            The PublicKey for the election
	 * @param rand
	 *            The random-number generator
	 * @throws IncorrectParameterException
	 */
	public Voter(int id, ElectionProperties ep, PublicKey pk, Random rand,
			BulletinBoard bb) {

		this.id = id;
		this.k = ep.getK();
		this.L = ep.getL();
		this.t = ep.getT();
		this.M = ep.getM();
		this.n = pk.getN();
		this.s = pk.getS();
		this.noOfVotes = ep.getNoOfVotes();
		this.rand = rand;
		this.voteValues = ep.getVoteValues();
		this.pk = pk;
		this.nArray = pk.getNArray();
		voteRandom = new BigInteger[ep.getNoOfVotes()];
		votePointer = 0;
		h = new HashFunction("SHA-256");
		precomp = preCompute();
		this.bb = bb;

		/*
		 * if((Utility.log2(M)*L) < ((double)(k-1)*s)){ throw new
		 * IncorrectParameterException("Invalid Parameters"); }
		 */

	}

	
/** Method for posting a vote to the bulletin board
 * 
 * @param candidate
 * @throws RemoteException
 * @throws IncorrectVoteException
 */
	public void postOneofLVote(int candidate) throws RemoteException,
			IncorrectVoteException {
		bb.postVote(encryptVote(candidate));
	}
	/** Method for posting a vote to the bulletin board
	 * 
	 * @param candidate Chosen candidates
	 * @throws RemoteException
	 * @throws IncorrectVoteException
	 */
	public void postKofLVote(int[] candidate) throws RemoteException,
			IncorrectVoteException {
		bb.postVote(encryptMultipleVotes(candidate));
	}
	/** Method for posting a vote to the bulletin board
	 * 
	 * @param candidate Chosen candidate
	 * @throws RemoteException
	 * @throws IncorrectVoteException
	 */
	public void postOneofLVoteBinary(int candidate) throws RemoteException,
			IncorrectVoteException {
		bb.postVote(encryptVoteBinary(candidate));
	}
	/** Method for posting a vote to the bulletin board
	 * 
	 * @param candidate Chosen candidates
	 * @throws RemoteException
	 * @throws IncorrectVoteException
	 */
	public void postKofLVoteBinary(int[] candidate) throws RemoteException,
			IncorrectVoteException {
		bb.postVote(encryptMultipleVoteBinary(candidate));
	}

	/**
	 * The method for encrypting a Vote and creating the proof of correctness in the case of 1-out-of-L elections
	 * 
	 * @param candidate
	 *            The candidate the Client voted for
	 * @return The Vote including the proof of correctness
	 * @throws IncorrectVoteException
	 */
	public Vote encryptVote(int candidate) throws IncorrectVoteException {

		if (candidate > L)
			throw new IncorrectVoteException("Choice out of bounds");

		BigInteger g = pk.getG();
		BigInteger z[] = new BigInteger[L];
		BigInteger a[] = new BigInteger[L];
		BigInteger e[] = new BigInteger[L];

		BigInteger valuesToBeHashed[] = new BigInteger[L + 2];
		valuesToBeHashed[0] = n;
		valuesToBeHashed[1] = BigInteger.valueOf(id);

		BigInteger r = null;
		BigInteger e_sum = BigInteger.ZERO;
		BigInteger pow_t = BigInteger.valueOf(2).pow(t);

		BigInteger[] encryption = encryptOptimized(candidate);
		BigInteger ciphertext_vote = encryption[0];
		voteRandom[votePointer] = encryption[1];
		BigInteger v = encryption[1];
		for (int i = 0; i < L; i++) {
			if (i == candidate) {
				// Invoke Protocol 1-out-of-k n^s'th power for the real vote.

				// Choose a random r in Z_n^(s+1) to use in a[i] = E(0,r)
				r = new BigInteger(n.bitCount(), rand);
				while (r.compareTo(n) >= 0
						|| r.gcd(n).compareTo(BigInteger.ONE) > 0)
					r = new BigInteger(n.bitCount(), rand);

				// Calculate a[i]
				a[i] = r.modPow(nArray[s], nArray[s + 1]);

			} else {
				// Invoke the honest-verifier simulator for n^s'th power
				// protocol

				// Choose z[i] random in Z_n rel.prime to n
				z[i] = new BigInteger(n.bitCount(), rand);
				while (z[i].compareTo(n) >= 0
						|| z[i].gcd(n).compareTo(BigInteger.ONE) > 0)
					z[i] = new BigInteger(n.bitCount(), rand);

				// Choose e[i] random in Z_(2^t)
				e[i] = new BigInteger(t, rand);
				e_sum = e_sum.add(e[i]).mod(pow_t);

				// Set a[i] =
				// (ciphertext_vote/g^votevalue[i])^(-e[i])*z[i]^(n^s) mod
				// n^(s+1)
				BigInteger temp = (g.modPow(voteValues[i], nArray[s + 1]))
						.modInverse(nArray[s + 1]);
				a[i] = (ciphertext_vote.multiply(temp).modPow(e[i],
						nArray[s + 1])).modInverse(nArray[s + 1]);
				a[i] = a[i].multiply(z[i].modPow(nArray[s], nArray[s + 1]));

			}
			// Include all a[i]-values into the hash-function-values
			valuesToBeHashed[i + 2] = a[i];

		}
		// Generate a hash-value as challenge, using Fiat-Shamir heuristic with
		// SHA-256
		BigInteger e_Challenge = h.generateHash(valuesToBeHashed, t);
		// Calculate the random challenge for e[candidate]
		e[candidate] = (e_Challenge.subtract(e_sum)).mod(pow_t);
		// Calculate z[candidate]
		z[candidate] = (r.multiply(v.modPow(e[candidate], nArray[s + 1])))
				.mod(nArray[s + 1]);

		return new Vote(ciphertext_vote, a, e_Challenge, z, e, id);
	}
/** Unoptimized encryption algorithm
 *  
 * @param candidate
 * @return The vote and the random value used in the computations
 */
	public BigInteger[] encrypt(int candidate) {

		BigInteger[] encryption = new BigInteger[2];

		BigInteger v = new BigInteger(n.bitCount(), rand);
		while (v.compareTo(n) >= 0 || v.gcd(n).compareTo(BigInteger.ONE) > 0)
			v = new BigInteger(n.bitCount(), rand);

		BigInteger temp1 = pk.getG().modPow(voteValues[candidate],
				nArray[s + 1]);
		BigInteger temp2 = v.modPow(nArray[s], nArray[s + 1]);
		BigInteger ciphertext_vote = (temp1.multiply(temp2)).mod(nArray[s + 1]);
		encryption[0] = ciphertext_vote;
		encryption[1] = v;

		return encryption;

	}
/** Method for precomputing values used in the optimized encryption algorithm
 * 
 * @return
 */
	private BigInteger[] preCompute() {
		BigInteger precomp[] = new BigInteger[s + 1];
		BigInteger jFac = BigInteger.ONE;
		precomp[1] = n;
		for (int j = 1; j <= s; j++) {
			jFac = jFac.multiply(BigInteger.valueOf(j));
			precomp[j] = jFac.modInverse(nArray[s + 1]);
			precomp[j] = precomp[j].multiply(n.pow(j)).mod(nArray[s + 1]);
		}
		return precomp;
	}
/** Optimized way of calculating (1+n)^exp
 * 
 * @param exp The Exponent for the caluculations
 * @return (1+n)^exp
 */
	private BigInteger calculateGM(BigInteger exp) {
		BigInteger c = BigInteger.ONE.add(exp.multiply(n));
		BigInteger temp = exp;
		for (int j = 2; j <= s; j++) {
			BigInteger j_ = BigInteger.valueOf(j);
			temp = temp.multiply((exp.subtract(j_)).add(BigInteger.ONE)).mod(
					nArray[s - j + 1]);
			c = (c.add(temp.multiply(precomp[j]))).mod(nArray[s + 1]);
		}
		return c;

	}
/** The optimized encryption algorithm
 * 
 * @param candidate
 * @return The ciphertext and the random value used to hide the plaintext
 */
	public BigInteger[] encryptOptimized(int candidate) {
		BigInteger[] encryption = new BigInteger[2];
		BigInteger voteValue = voteValues[candidate];
		BigInteger c = BigInteger.ONE.add(voteValue.multiply(n));
		BigInteger temp = voteValue;
		for (int j = 2; j <= s; j++) {
			BigInteger j_ = BigInteger.valueOf(j);
			temp = temp.multiply((voteValue.subtract(j_)).add(BigInteger.ONE))
					.mod(nArray[s - j + 1]);
			c = (c.add(temp.multiply(precomp[j]))).mod(nArray[s + 1]);
		}
		BigInteger v = new BigInteger(n.bitCount(), rand);
		while (v.compareTo(n) >= 0 || v.gcd(n).compareTo(BigInteger.ONE) > 0)
			v = new BigInteger(n.bitCount(), rand);

		BigInteger rns = v.modPow(nArray[s], nArray[s + 1]);
		encryption[0] = c.multiply(rns).mod(nArray[s + 1]);
		encryption[1] = v;

		return encryption;
	}
/** Optimized encryption algorithm for encrypting a BigInteger. Mainly used when making the zero-knowledge proofs.
 * 
 * @param value
 * @return The ciphertext and the random value used to hide the plaintext
 */
	public BigInteger[] encryptOptimized(BigInteger value) {
		BigInteger[] encryption = new BigInteger[2];
		BigInteger c = BigInteger.ONE.add(value.multiply(n));
		BigInteger temp = value;
		for (int j = 2; j <= s; j++) {
			BigInteger j_ = BigInteger.valueOf(j);
			temp = temp.multiply((value.subtract(j_)).add(BigInteger.ONE)).mod(
					nArray[s - j + 1]);
			c = (c.add(temp.multiply(precomp[j]))).mod(nArray[s + 1]);
		}
		BigInteger v = new BigInteger(n.bitCount(), rand);
		while (v.compareTo(n) >= 0 || v.gcd(n).compareTo(BigInteger.ONE) > 0)
			v = new BigInteger(n.bitCount(), rand);

		BigInteger rns = v.modPow(nArray[s], nArray[s + 1]);
		encryption[0] = c.multiply(rns).mod(nArray[s + 1]);
		encryption[1] = v;

		return encryption;
	}

/** Encrypts the chosen votes using the method explained as k-out-of-L voting.
 * 
 * @param candidate chosen candidates
 * @return The MultipleVotes object containing the votes and proofs needed.
 * @throws IncorrectVoteException If any choices were out of bounds, or if incorrect number of votes were cast.
 */

	public MultipleVotes encryptMultipleVotes(int[] candidate)
			throws IncorrectVoteException {

		if (candidate.length != noOfVotes)
			throw new IncorrectVoteException("Incorrect number of votes cast");

		Vote[] votes = new Vote[candidate.length];

		for (; votePointer < candidate.length; votePointer++) {

			votes[votePointer] = encryptVote(candidate[votePointer]);

		}
		// Public input: n, g, E(a), E(b), E(c) where a*b = c mod n
		BigInteger ea[][] = new BigInteger[candidate.length][candidate.length];
		BigInteger eb[][] = new BigInteger[candidate.length][candidate.length];
		BigInteger ec[][] = new BigInteger[candidate.length][candidate.length];
		BigInteger ra[][] = new BigInteger[candidate.length][candidate.length];
		BigInteger rb[][] = new BigInteger[candidate.length][candidate.length];
		BigInteger rc[][] = new BigInteger[candidate.length][candidate.length];
		BigInteger A[][] = new BigInteger[candidate.length][candidate.length];
		BigInteger B[][] = new BigInteger[candidate.length][candidate.length];

		for (int i = 0; i < candidate.length; i++) {
			for (int j = i + 1; j < candidate.length; j++) {
				A[i][j] = voteValues[candidate[i]]
						.subtract(voteValues[candidate[j]]);
				B[i][j] = A[i][j].modInverse(nArray[s + 1]);
				BigInteger[] temp = encryptOptimized(A[i][j]);
				ea[i][j] = (votes[i].getVote().multiply(votes[j].getVote()
						.modInverse(nArray[s + 1]))).mod(nArray[s + 1]);// temp[0];
				ra[i][j] = (voteRandom[i].multiply(voteRandom[j]
						.modInverse(nArray[s + 1]))).mod(nArray[s + 1]);// temp[1];

				temp = encryptOptimized(B[i][j]);
				eb[i][j] = temp[0];
				rb[i][j] = temp[1];

				temp = encryptOptimized(BigInteger.ONE);
				ec[i][j] = temp[0];
				rc[i][j] = temp[1];

			}
		}

		BigInteger[][] d = new BigInteger[candidate.length][candidate.length];
		BigInteger[][] rd = new BigInteger[candidate.length][candidate.length];
		BigInteger[][] ed = new BigInteger[candidate.length][candidate.length];
		BigInteger[][] edb = new BigInteger[candidate.length][candidate.length];
		BigInteger[][] rdb = new BigInteger[candidate.length][candidate.length];
		BigInteger[][] e = new BigInteger[candidate.length][candidate.length];
		BigInteger[][] f = new BigInteger[candidate.length][candidate.length];
		BigInteger[][] z1 = new BigInteger[candidate.length][candidate.length];
		BigInteger[][] z2 = new BigInteger[candidate.length][candidate.length];

		for (int i = 0; i < candidate.length; i++) {
			for (int j = i + 1; j < candidate.length; j++) {
				d[i][j] = new BigInteger(nArray[s].bitLength(), rand);
				BigInteger[] temp2 = encryptOptimized(d[i][j]);
				ed[i][j] = temp2[0];
				rd[i][j] = temp2[1];
				temp2 = encryptOptimized(d[i][j].multiply(B[i][j]));
				edb[i][j] = temp2[0];
				rdb[i][j] = temp2[1];

				BigInteger[] valuesToBeHashed = new BigInteger[5];
				valuesToBeHashed[0] = n;
				valuesToBeHashed[1] = ea[i][j];
				valuesToBeHashed[2] = eb[i][j];
				valuesToBeHashed[3] = ec[i][j];
				valuesToBeHashed[4] = BigInteger.valueOf(id);
				e[i][j] = h.generateHash(valuesToBeHashed, t);
				f[i][j] = (e[i][j].multiply(A[i][j])).add(d[i][j]).mod(
						nArray[s]);

				z1[i][j] = ((ra[i][j].modPow(e[i][j], n)).multiply(rd[i][j]))
						.mod(n);
				z2[i][j] = rdb[i][j].multiply(rc[i][j].modPow(e[i][j], n));
				z2[i][j] = z2[i][j].modInverse(n);
				z2[i][j] = (z2[i][j].multiply(rb[i][j].modPow(f[i][j],
						nArray[s + 1]))).mod(n);
			}
		}
		MultipleVotes mVotes = new MultipleVotes(votes, ea, eb, ec, ed, edb, e,
				f, z1, z2, id);

		return mVotes;
	}
	
	/** Encrypts the chosen votes using the method explained as k-out-of-L Binary encoded voting.
	 * 
	 * @param candidate chosen candidates
	 * @return The MultipleBinaryVotes object containing the votes and proofs needed.
	 * @throws IncorrectVoteException If any choices were out of bounds, or if incorrect number of votes were cast.
	 */
	public MultipleBinaryVotes encryptMultipleVoteBinary(int candidate[])
			throws IncorrectVoteException {
		BigInteger M_ = BigInteger.valueOf(M);
		if (candidate.length != noOfVotes)
			throw new IncorrectVoteException("Incorrect number of votes cast");
		for (int i = 0; i < candidate.length; i++)
			System.out.println(candidate[i]);
		BinaryVote[] votes = new BinaryVote[candidate.length];

		for (votePointer = 0; votePointer < candidate.length; votePointer++) {
			votes[votePointer] = encryptVoteBinary(candidate[votePointer]);
		}
		// Public input: n, g, E(a), E(b), E(c) where a*b = c mod n
		BigInteger ea[][] = new BigInteger[candidate.length][candidate.length];
		BigInteger eb[][] = new BigInteger[candidate.length][candidate.length];
		BigInteger ec[][] = new BigInteger[candidate.length][candidate.length];
		BigInteger ra[][] = new BigInteger[candidate.length][candidate.length];
		BigInteger rb[][] = new BigInteger[candidate.length][candidate.length];
		BigInteger rc[][] = new BigInteger[candidate.length][candidate.length];
		BigInteger A[][] = new BigInteger[candidate.length][candidate.length];
		BigInteger B[][] = new BigInteger[candidate.length][candidate.length];

		for (int i = 0; i < candidate.length; i++) {
			for (int j = i + 1; j < candidate.length; j++) {

				A[i][j] = ((M_.modPow(BigInteger.valueOf(candidate[i]),
						nArray[s + 1])).subtract(M_.modPow(BigInteger
						.valueOf(candidate[j]), nArray[s + 1])));

				B[i][j] = A[i][j].modInverse(nArray[s + 1]).mod(nArray[s + 1]);
				BigInteger[] temp = encryptOptimized(A[i][j]);


				BigInteger temp2 = votes[j].getFi()[l]
						.modInverse(nArray[s + 1]);
				ea[i][j] = ((votes[i].getFi()[l]).multiply(temp2))
						.mod(nArray[s + 1]);
				ra[i][j] = (voteRandom[i].multiply(voteRandom[j]
						.modInverse(nArray[s + 1]))).mod(nArray[s + 1]);// temp[1];

				temp = encryptOptimized(B[i][j]);
				eb[i][j] = temp[0];
				rb[i][j] = temp[1];

				temp = encryptOptimized(BigInteger.ONE);
				ec[i][j] = temp[0];
				rc[i][j] = temp[1];

			}
		}

		BigInteger[][] d = new BigInteger[candidate.length][candidate.length];
		BigInteger[][] rd = new BigInteger[candidate.length][candidate.length];
		BigInteger[][] ed = new BigInteger[candidate.length][candidate.length];
		BigInteger[][] edb = new BigInteger[candidate.length][candidate.length];
		BigInteger[][] rdb = new BigInteger[candidate.length][candidate.length];
		BigInteger[][] e = new BigInteger[candidate.length][candidate.length];
		BigInteger[][] f = new BigInteger[candidate.length][candidate.length];
		BigInteger[][] z1 = new BigInteger[candidate.length][candidate.length];
		BigInteger[][] z2 = new BigInteger[candidate.length][candidate.length];

		for (int i = 0; i < candidate.length; i++) {
			for (int j = i + 1; j < candidate.length; j++) {
				d[i][j] = new BigInteger(nArray[s].bitLength(), rand);
				BigInteger[] temp2 = encryptOptimized(d[i][j]);
				ed[i][j] = temp2[0];
				rd[i][j] = temp2[1];
				temp2 = encryptOptimized(d[i][j].multiply(B[i][j]));
				edb[i][j] = temp2[0];
				rdb[i][j] = temp2[1];

				BigInteger[] valuesToBeHashed = new BigInteger[5];
				valuesToBeHashed[0] = n;
				valuesToBeHashed[1] = ea[i][j];
				valuesToBeHashed[2] = eb[i][j];
				valuesToBeHashed[3] = ec[i][j];
				valuesToBeHashed[4] = BigInteger.valueOf(id);
				e[i][j] = h.generateHash(valuesToBeHashed, t);

				f[i][j] = (e[i][j].multiply(A[i][j])).add(d[i][j]).mod(
						nArray[s]);

				z1[i][j] = ((ra[i][j].modPow(e[i][j], n)).multiply(rd[i][j]))
						.mod(n);
				z2[i][j] = rdb[i][j].multiply(rc[i][j].modPow(e[i][j], n));
				z2[i][j] = z2[i][j].modInverse(n);
				z2[i][j] = (z2[i][j].multiply(rb[i][j].modPow(f[i][j],
						nArray[s + 1]))).mod(n);
			}
		}
		MultipleBinaryVotes mVotes = new MultipleBinaryVotes(votes, ea, eb, ec,
				ed, edb, e, f, z1, z2, id);

		return mVotes;

	}
	/** Encrypts a vote in the case of 1-out-of-L Binary Voting
	 * 
	 * @param candidate The candidate that was voted for
	 * @return The BinaryVote holding the ciphertext and the needed zero-knowledge proofs of correctness
	 * @throws IncorrectVoteException If the choice was out of bounds
	 */
	public BinaryVote encryptVoteBinary(int candidate)
			throws IncorrectVoteException {
		if (candidate > L)
			throw new IncorrectVoteException("Choice out of bounds");

		l = 1;
		BigInteger M_ = BigInteger.valueOf(M);
		while (Math.pow(2, l + 1) < L)
			l++;

		String bitRep = Integer.toString(candidate, 2);

		BigInteger g = pk.getG();
		BigInteger pow_t = BigInteger.valueOf(2).pow(t);
		BigInteger encryptions[] = new BigInteger[l + 1];
		BigInteger rb[] = new BigInteger[l + 1];
		BigInteger[][] a = new BigInteger[2][l + 1];
		BigInteger[][] e = new BigInteger[2][l + 1];
		BigInteger[][] z = new BigInteger[2][l + 1];

		BigInteger valuesToBeHashed[] = new BigInteger[4];
		valuesToBeHashed[0] = n;
		valuesToBeHashed[1] = BigInteger.valueOf(id);
		BigInteger MraisedToBitValue[] = new BigInteger[l + 1];

		for (int i = 0; i <= l; i++) {
			int biti = 0;
			try {
				biti = candidate % 2;
				candidate = candidate / 2;
			} catch (Exception ex) {
				biti = 0;
			}
			if (biti == 1) {
				// The bit in question is 1. Create encryption and proofs for
				// this case.
				// System.out.println(biti);
				MraisedToBitValue[i] = M_.pow((int) Math.pow(2, i));
				BigInteger[] tempVote = encryptOptimized(MraisedToBitValue[i]);
				encryptions[i] = tempVote[0];
				BigInteger v = tempVote[1];
				rb[i] = v;
				// Choose a random r in Z_n^(s+1) to use in a[i] = E(0,r)
				BigInteger r = new BigInteger(n.bitCount(), rand);
				while (r.compareTo(n) >= 0
						|| r.gcd(n).compareTo(BigInteger.ONE) > 0)
					r = new BigInteger(n.bitCount(), rand);

				// Calculate a[i]
				a[0][i] = r.modPow(nArray[s], nArray[s + 1]);

				// Create honest-verifier simulator for the other value.
				// Choose z[i] random in Z_n rel.prime to n
				z[1][i] = new BigInteger(n.bitCount(), rand);
				while (z[1][i].compareTo(n) >= 0
						|| z[1][i].gcd(n).compareTo(BigInteger.ONE) > 0)
					z[1][i] = new BigInteger(n.bitCount(), rand);

				// Choose e[i] random in Z_(2^t)
				e[1][i] = new BigInteger(t, rand);

				// Set a[i] =
				// (ciphertext_vote/g^votevalue[i])^(-e[i])*z[i]^(n^s) mod
				// n^(s+1)
				BigInteger temp = g.modInverse(nArray[s + 1]);
				a[1][i] = (encryptions[i].multiply(temp).modPow(e[1][i],
						nArray[s + 1])).modInverse(nArray[s + 1]);
				a[1][i] = a[1][i].multiply(z[1][i].modPow(nArray[s],
						nArray[s + 1]));

				valuesToBeHashed[2] = a[0][i];
				valuesToBeHashed[3] = a[1][i];

				BigInteger e_Challenge = h.generateHash(valuesToBeHashed, t)
						.mod(pow_t);

				e[0][i] = e_Challenge.subtract(e[1][i]).mod(pow_t);
				;
				z[0][i] = (r.multiply(v.modPow(e[0][i], nArray[s + 1])))
						.mod(nArray[s + 1]);
			} else {

				// The bit in question is 0. Create encryption and proofs for
				// this case.
				MraisedToBitValue[i] = BigInteger.ONE;
				BigInteger[] tempVote = encryptOptimized(BigInteger.ONE);
				encryptions[i] = tempVote[0];
				BigInteger v = tempVote[1];
				rb[i] = v;

				// Choose a random r in Z_n to use in a[i] = E(0,r)
				BigInteger r = new BigInteger(n.bitCount(), rand);
				while (r.compareTo(n) >= 0
						|| r.gcd(n).compareTo(BigInteger.ONE) > 0)
					r = new BigInteger(n.bitCount(), rand);

				// Calculate a[i]
				a[1][i] = r.modPow(nArray[s], nArray[s + 1]);

				// Create honest-verifier simulator for the other value.
				// Choose z[i] random in Z_n rel.prime to n
				z[0][i] = new BigInteger(n.bitCount(), rand);
				while (z[0][i].compareTo(n) >= 0
						|| z[0][i].gcd(n).compareTo(BigInteger.ONE) > 0)
					z[0][i] = new BigInteger(n.bitCount(), rand);

				// Choose e[i] random in Z_(2^t)
				e[0][i] = new BigInteger(t, rand);

				// Set a[i] = (ciphertext_vote/g^(M^2^i))^(-e[i])*z[i]^(n^s) mod
				// n^(s+1)
				BigInteger temp = (calculateGM(M_.pow((int) Math.pow(2, i))))
						.modInverse(nArray[s + 1]);
				a[0][i] = (encryptions[i].multiply(temp).modPow(e[0][i],
						nArray[s + 1])).modInverse(nArray[s + 1]);
				a[0][i] = a[0][i].multiply(z[0][i].modPow(nArray[s],
						nArray[s + 1]));

				valuesToBeHashed[2] = a[0][i];
				valuesToBeHashed[3] = a[1][i];

				BigInteger e_Challenge = h.generateHash(valuesToBeHashed, t)
						.mod(pow_t);
				e[1][i] = e_Challenge.subtract(e[0][i]).mod(pow_t);
				z[1][i] = (r.multiply(v.modPow(e[1][i], nArray[s + 1])))
						.mod(nArray[s + 1]);

			}
		}

		// Proofs for that bits are 1 or 0 are created. Now to the
		// Proofs for the 2nd part
		BigInteger Fi[] = new BigInteger[l + 1];
		BigInteger fi[] = new BigInteger[l + 1];
		BigInteger rfi[] = new BigInteger[l + 1];

		Fi[0] = MraisedToBitValue[0];
		fi[0] = encryptions[0];
		rfi[0] = rb[0];
		
		BigInteger[] tempVote;
		for (int i = 1; i <= l; i++) {
			Fi[i] = Fi[i - 1].multiply(MraisedToBitValue[i]);
			 tempVote = encryptOptimized(Fi[i]);
			fi[i] = tempVote[0];
			rfi[i] = tempVote[1];
			if (i == l) {
				voteRandom[votePointer] = rfi[i];
			}
		}

		// Make the multiplication proof!
		BigInteger z1[] = new BigInteger[l];
		BigInteger z2[] = new BigInteger[l];
		BigInteger d[] = new BigInteger[l];
		BigInteger ed[] = new BigInteger[l];
		BigInteger edb[] = new BigInteger[l];
		BigInteger rdb[] = new BigInteger[l];
		BigInteger rd[] = new BigInteger[l];
		BigInteger eMult[] = new BigInteger[l];
		BigInteger f[] = new BigInteger[l];
		BigInteger[] valuesToBeHashed2 = new BigInteger[5];
		valuesToBeHashed2[0] = n;
		valuesToBeHashed2[4] = BigInteger.valueOf(id);
		for (int i = 0; i < l; i++) {
			d[i] = new BigInteger(nArray[s].bitLength(), rand);
			BigInteger[] temp2 = encryptOptimized(d[i]);
			ed[i] = temp2[0];
			rd[i] = temp2[1];
			temp2 = encryptOptimized(d[i].multiply(encryptions[i + 1]));
			edb[i] = temp2[0];
			rdb[i] = temp2[1];

			valuesToBeHashed2[1] = fi[i];
			valuesToBeHashed2[2] = encryptions[i + 1];
			valuesToBeHashed2[3] = fi[i + 1];

			eMult[i] = h.generateHash(valuesToBeHashed2, t);

			f[i] = (eMult[i].multiply(Fi[i])).add(d[i]).mod(nArray[s]);
			z1[i] = ((rfi[i].modPow(eMult[i], n)).multiply(rd[i])).mod(n);
			z2[i] = rdb[i].multiply(rfi[i + 1].modPow(eMult[i], n));
			z2[i] = z2[i].modInverse(nArray[s + 1]);
			z2[i] = (z2[i].multiply(rb[i + 1].modPow(f[i], nArray[s + 1])))
					.mod(n);

		}

		return new BinaryVote(encryptions, e, z, a, fi, z1, z2, eMult, ed, edb,
				f, id);

	}

}
