package votingSystem;

import java.io.Serializable;
import java.math.BigInteger;

/**
 * Class to hold the all parameters of an election
 * 
 * @author Bent K, Onshus
 * 
 */
public class ElectionProperties implements Serializable {
	private int[] voteWeights;

	private int k, L, M, l, w, t, s, electionType;

	private BigInteger[] voteValues;

	private int noOfVotes;

	private String[] voteNames;
/** The constructor giving all parameters needed to specify an election, as well as calculating the s to be used in the election.
 * 
 * @param k The bitlength of the key
 * @param L The number of candidates
 * @param M The maximum number of voters
 * @param l The number of decryption servers
 * @param w The number of valid KeyShares needed for decrypting the result
 * @param t The bitlength of the challenge used in the zero-knowledge proofs
 * @param s The exponent used in the cryptosystem
 * @param noOfVotes Number of votes each voter is given
 * @param voteNames The names of the candidates
 * @param electionType The type of election.  0 if 1-out of L, 1 if k-out-of-L, 2 if 1-out-of-L Binary, 3 if k-out-of-L Binary.
 * @param voteWeights The weight of each vote
 */
	public ElectionProperties(int k, int L, int M, int l, int w, int t, int noOfVotes, String[] voteNames, int electionType,
			int[] voteWeights) {

		this.k = k;
		this.L = L;
		this.M = M;
		this.l = l;
		this.w = w;
		this.t = t;
		this.s = 1;
		this.voteWeights = voteWeights;
		this.electionType = electionType;
		this.noOfVotes = noOfVotes;
		this.voteNames = voteNames;
		voteValues = generateVoteValues(L, M);

		//Find s to use in the election if larger s than 1 is needed. If electiontype 1-out-of-L or k-out-of-L
		// a message space of n^s>M^L is needed. if Binary votes are used, the message space should be M^(2^(l+1)) <n^s
		// is needed. where l is the bitlength of the binary rep. of L.
		if(electionType ==0 || electionType ==1){
			while(L*(Utility.log2(M)) >= (k-1)*s)
				s++;			
		}
		else{
			int l_bit=1;
			while (Math.pow(2, l_bit + 1) < L)
				l_bit++;
			
			while(L*Utility.log2((int)Math.pow(2,l_bit+1)) > (k-1)*s)
				s++;
							
		}
		
		System.out.println("Parameter s = "+s);
	}

	/**
	 * Generates the values needed to vote for candidates.
	 * 
	 * @param L
	 * @param M
	 * @return The values used for votes for all candidates
	 */
	public static BigInteger[] generateVoteValues(int L, int M) {
		BigInteger[] voteValues = new BigInteger[L];
		voteValues[0] = BigInteger.valueOf(M);
		for (int i = 1; i < L; i++) {
			voteValues[i] = voteValues[i - 1].multiply(BigInteger.valueOf(M));
		}

		return voteValues;

	}

	/**
	 * 
	 * @return The bitlength of n
	 */
	public int getK() {
		return k;
	}

	/**
	 * 
	 * @return Number of decryption servers
	 */
	public int getl() {
		return l;
	}

	/**
	 * 
	 * @return Number of candidates in election
	 */
	public int getL() {
		return L;
	}

	/**
	 * 
	 * @return Maximum number of voters in the election
	 */
	public int getM() {
		return M;
	}

	/**
	 * 
	 * @return n^s
	 */
	public int getS() {
		return s;
	}

	/**
	 * 
	 * @return Bitlength of challenges
	 */
	public int getT() {
		return t;
	}

	/**
	 * 
	 * @return Number of decryption shares needed to decrypt result.
	 */
	public int getW() {
		return w;
	}

	/**
	 * 
	 * @return The values needed to vote for each candidate.
	 */
	public BigInteger[] getVoteValues() {
		return voteValues;
	}

	public int getNoOfVotes() {
		return noOfVotes;
	}

	public String[] getVoteNames() {
		return voteNames;
	}

	public int getElectionType() {
		return electionType;
	}

	public int[] getVoteWeights() {
		return voteWeights;
	}
}
